-- API: PRIVATE
create function docker_api_base_url()
    returns text
as
'MODULE_PATHNAME',
'docker_api_base_url'
    language c;

-- API: PRIVATE
create function docker_host_ip()
    returns text
as
'MODULE_PATHNAME',
'docker_host_ip'
    language c;

-- API: PRIVATE
create function docker_images_json()
    returns jsonb
as
$$
declare
    response omni_httpc.http_response;
begin
    select *
    into response
    from
        omni_httpc.http_execute(
                omni_httpc.http_request(
                        format('http://%s/images/json', omni_containers.docker_api_base_url())
                    )
            );
    if response.status = 200 then
        return convert_from(response.body, 'UTF8');
    else
        raise exception 'Can''t list docker images' using
            detail = format('Error code %s: %s', response.status, convert_from(response.body, 'UTF8')::jsonb);
    end if;
end;
$$ language plpgsql;

comment
    on function docker_images_json() is 'Private API';

-- API: PUBLIC
create view docker_images as
    (
    select
        "Id"                    as id,
        "Size"                  as size,
        "Labels"                as labels,
        to_timestamp("Created") as created_at,
        "ParentId"              as parent_id,
        "RepoTags"              as repo_tags,
        "Containers"            as containers,
        "SharedSize"            as shared_size,
        "RepoDigests"           as repo_digests,
        "VirtualSize"           as virtual_size
    from
        jsonb_to_recordset(jsonb_strip_nulls(omni_containers.docker_images_json()))
            as images("Id" text, "Size" int8, "Labels" jsonb, "Created" int8, "ParentId" text, "RepoTags" text[],
                      "Containers" int, "SharedSize" int, "RepoDigests" text[], "VirtualSize" jsonb)
        );

-- API: PUBLIC
create function docker_container_create(
    image text,
    cmd text default null,
    attach text default 'db.omni',
    start bool default true,
    wait bool default false,
    pull bool default false,
    options jsonb default '{}')
    returns text
as
$$
declare
    normalized_image text;
    post_body jsonb        = options;
    conn_info jsonb        = '[]';
    port text              = '5432';
    retry_creating bool    = false;
    attempted_to_pull bool = false;
    response omni_httpc.http_response;
    container_id text;
begin
    case (length(image) - length(replace(image, '/', '')))
        when 0 then -- No slashes, it's a library
        normalized_image = 'docker.io/library/' || image;
        when 1 then -- One slash, it's on docker.io
        normalized_image = 'docker.io/' || image;
        else -- More than one slash, it's a normalized URl
        normalized_image = image;
        end case;

    post_body = post_body || jsonb_build_object('Image', normalized_image);

    if cmd is not null then
        post_body = post_body || jsonb_build_object('Cmd', jsonb_build_array('sh', '-c', cmd));
    end if;

    if attach is not null then
        conn_info = conn_info::jsonb || to_jsonb(format('PGUSER=%s', current_user));
        conn_info = conn_info::jsonb || to_jsonb(format('PGHOST=%s', attach));
        conn_info = conn_info::jsonb || to_jsonb(format('PGDATABASE=%s', current_database()));
        select setting into port from pg_settings where name = 'port';
        conn_info = conn_info::jsonb || to_jsonb(format('PGPORT=%s', port));

        post_body = jsonb_set(post_body, '{Env}',
                              coalesce(post_body -> 'Env', '[]')::jsonb || conn_info);

        if post_body -> 'HostConfig' is null then
            post_body = jsonb_set(post_body, '{HostConfig}', '{
              "ExtraHosts": []
            }');
        end if;

        if omni_containers.docker_host_ip() is null then
            post_body = jsonb_set(post_body, '{HostConfig,ExtraHosts}',
                                  coalesce(post_body -> 'HostConfig' -> 'ExtraHosts', '[]')::jsonb ||
                                  to_jsonb(format('%s:host-gateway',
                                                  attach))
                );
        else
            post_body = jsonb_set(post_body, '{HostConfig,ExtraHosts}',
                                  coalesce(post_body -> 'HostConfig' -> 'ExtraHosts', '[]')::jsonb ||
                                  to_jsonb(format('%s:%s',
                                                  attach, omni_containers.docker_host_ip()))
                );
        end if;
    end if;

    loop
        retry_creating = false;
        -- Attempt to create the container
        select *
        into response
        from
            omni_httpc.http_execute(omni_httpc.http_request(
                    format('http://%s/containers/create', omni_containers.docker_api_base_url()),
                    'POST',
                    array [
                        omni_http.http_header('Content-Type', 'application/json')
                        ],
                    convert_to(post_body::text, 'UTF8')
                ));

        case
            when response.status = 201 then -- We're done
            container_id = convert_from(response.body, 'UTF8')::jsonb ->> 'Id';
            when response.status = 404 then -- Not found
            if pull and not attempted_to_pull then -- Try to pull the image if allowed to
                declare
                    url text;
                begin
                raise notice 'Pulling image %', image;
                url := 'http://%s/images/create?fromImage=%s';
                if normalized_image not like '%:%' then
                    url := url || '&tag=latest';
                end if;
                select *
                into response
                from
                    omni_httpc.http_execute(omni_httpc.http_request(
                            format(
                                    url,
                                    omni_containers.docker_api_base_url(),
                                    omni_web.url_encode(normalized_image)
                                ),
                            'POST')
                        );
                attempted_to_pull = true;

                if response.status = 200 then
                    retry_creating = true; -- Try to create the container again
                    continue;
                elsif response.status != 200 then
                    raise exception 'Failed to pull image %', image using
                        detail = format('Error code %s: %s', response.status, convert_from(response.body, 'UTF8'));
                end if;
                end;
            else
                raise exception 'Docker image not found' using detail = format('%s', image); -- Attempted, wasn't found
            end if;
            else -- Other error
            raise exception 'Can''t create the container' using
                detail = format('Error code %s: %s', response.status, convert_from(response.body, 'UTF8'));
            end case;
        exit when not retry_creating;
    end loop;

    if start = true then
        select *
        into response
        from
            omni_httpc.http_execute(omni_httpc.http_request(
                    format('http://%s/containers/%s/start', omni_containers.docker_api_base_url(), container_id),
                    'POST'
                ));
        if response.status != 204 then
            raise exception 'Can''t start the container' using
                detail = format('Error code %s: %s', response.status, response.headers);
        end if;
    end if;

    if wait = true then
        select *
        into response
        from
            omni_httpc.http_execute(omni_httpc.http_request(
                    format('http://%s/containers/%s/wait', omni_containers.docker_api_base_url(), container_id),
                    'POST'
                ));
        if response.status != 200 then
            raise exception 'Can''t wait for the container' using
                detail = format('Error code %s: %s', response.status, convert_from(response.body, 'UTF8'));
        end if;
    end if;

    return container_id;
end;
$$ language plpgsql;

-- API: PUBLIC
create function docker_container_inspect(id text)
    returns jsonb
as
$$
declare
response omni_httpc.http_response;
begin
    select * into response from omni_httpc.http_execute(
        omni_httpc.http_request(
            format('http://%s/containers/%s/json', omni_containers.docker_api_base_url(), id)
        )
    );
    case response.status
        when 200 then
            return convert_from(response.body, 'UTF8')::jsonb;
        else
            raise exception 'Can''t inspect the container' using
                detail = format('Error code %s: %s', response.status, convert_from(response.body, 'UTF8')::jsonb);
    end case;
end;
$$ language plpgsql;

-- API: PRIVATE
create function docker_stream_to_text(
    stream bytea
)
    returns text
as
'MODULE_PATHNAME',
'docker_stream_to_text'
    language c;


-- API: PUBLIC
create function docker_container_logs(
    id text,
    stdout bool default true,
    stderr bool default true,
    since timestamp default null,
    until timestamp default null,
    timestamps bool default false,
    tail int default null
)
    returns text
as
$$
declare
    query_params text = '';
    response     omni_httpc.http_response;
begin
    if stdout = true then
        query_params = query_params || '&stdout=true';
    else
        query_params = query_params || '&stdout=false';
    end if;

    if stderr = true then
        query_params = query_params || '&stderr=true';
    else
        query_params = query_params || '&stderr=false';
    end if;

    if since is not null then
        query_params = query_params || format('&since=%s', extract(epoch from since)::bigint);
    end if;

    if until is not null then
        query_params = query_params || format('&until=%s', extract(epoch from until)::bigint);
    end if;

    if timestamps = true then
        query_params = query_params || '&timestamps=true';
    else
        query_params = query_params || '&timestamps=false';
    end if;

    if tail is null or tail = -1 then
        query_params = query_params || '&tail=all';
    else
        query_params = query_params || format('&tail=%s', tail);
    end if;

    select *
    into response
    from
        omni_httpc.http_execute(
                omni_httpc.http_request(
                            format('http://%s/containers/%s/logs?', omni_containers.docker_api_base_url(), id) ||
                            query_params
                    )
            );

    if response.status = 200 then
        return omni_containers.docker_stream_to_text(response.body);
    else
        raise exception 'Can''t get logs from the container' using
            detail = format('Error code %s: %s', response.status, convert_from(response.body, 'UTF8'));
    end if;
end;
$$ language plpgsql;
-- API: PUBLIC
create function docker_container_stop(id text, timeout int default null)
    returns void
as
$$
declare
    response omni_httpc.http_response;
begin
    select *
    into response
    from
        omni_httpc.http_execute_with_options(
                omni_httpc.http_execute_options(timeout => timeout, first_byte_timeout => timeout),
                omni_httpc.http_request(
                        format('http://%s/containers/%s/stop', omni_containers.docker_api_base_url(), id),
                        'POST'
                    )
            );
    case response.status
        when 204 then -- We're done
        return;
        when 304 then raise exception 'Can''t stop the container' using
            detail = 'Container already stopped';
        when 404 then raise exception 'Can''t stop the container' using
            detail = 'No such container';
        else raise exception 'Can''t stop the container: % (status % error: %)', convert_from(response.body, 'utf8'), response.status, response.error;
        end case;
end;
$$ language plpgsql;
-- API: PUBLIC
create function docker_container_exec(id text, cmd text,
                                      attach_stdout boolean default false,
                                      attach_stderr boolean default false,
                                      options jsonb default '{}')
    returns text
as
$$
declare
    response  omni_httpc.http_response;
    post_body jsonb = options;
begin
    if cmd is not null then
        post_body = post_body || jsonb_build_object('Cmd', jsonb_build_array('sh', '-c', cmd));
    end if;
    if attach_stdout then
        post_body = post_body || jsonb_build_object('AttachStdout', true);
    end if;
    if attach_stderr then
        post_body = post_body || jsonb_build_object('AttachStderr', true);
    end if;

    select *
    into response
    from
        omni_httpc.http_execute_with_options(
                omni_httpc.http_execute_options(),
                omni_httpc.http_request(
                        format('http://%s/containers/%s/exec', omni_containers.docker_api_base_url(), id),
                        'POST',
                        array [
                            omni_http.http_header('Content-Type', 'application/json')
                            ],
                        convert_to(post_body::text, 'UTF8')
                )
        );
    case response.status
        when 201 then -- it's prepared
        null;
        when 404 then raise exception 'Can''t exec in the container' using
            detail = 'No such container';
        else raise exception 'Can''t exec in the container (HTTP %)', response.status using detail = (convert_from(response.body, 'utf8')::json) ->> 'message';
        end case;
    id := (convert_from(response.body, 'utf8')::jsonb ->> 'Id');

    post_body := '{}';
    if attach_stdout or attach_stderr then
        post_body = post_body || jsonb_build_object('Detach', false);
    end if;
    select *
    into response
    from
        omni_httpc.http_execute_with_options(
                omni_httpc.http_execute_options(),
                omni_httpc.http_request(
                        format('http://%s/exec/%s/start', omni_containers.docker_api_base_url(), id),
                        'POST',
                        array [
                            omni_http.http_header('Content-Type', 'application/json')
                            ],
                        convert_to(post_body::text, 'UTF8')
                )
        );
    case response.status
        when 200 then -- it's prepared
        if attach_stdout or attach_stderr then
            return omni_containers.docker_stream_to_text(response.body);
        else
            return null;
        end if;
        when 404 then raise exception 'Can''t exec in the container' using
            detail = 'No such container';
        else raise exception 'Can''t exec in the container (HTTP %)', response.status using detail = (convert_from(response.body, 'utf8')::json) ->> 'message';
        end case;
end;
$$ language plpgsql;
