create table languages
(
    id                       integer primary key generated always as identity,
    file_extension           text not null unique,
    language                 name not null,
    extension                name,
    file_processor           name,
    file_processor_extension name,
    priority                 integer not null default 0
);

insert
into
    languages (file_extension, language, extension, file_processor, file_processor_extension)
values
    ('sql', 'sql', null, null, null),
    ('pl', 'plperlu', 'plperlu', null, null),
    ('trusted.pl', 'plperl', 'plperl', null, null),
    ('py', 'plpython3u', 'plpython3u', 'omni_python.create_functions', 'omni_python'),
    ('tcl', 'pltclu', 'pltclu', null, null),
    ('trusted.tcl', 'pltcl', 'pltcl', null, null),
    ('rs', 'plrust', 'plrust', null, null);


select pg_catalog.pg_extension_config_dump('languages', '');
create table auxiliary_tools
(
    id                  integer primary key generated always as identity,
    filename_stem       text,
    filename_extension  text,
    processor           name not null,
    processor_extension name,
    priority            integer not null default 0
);

insert
into
    auxiliary_tools (filename_stem, filename_extension, processor, processor_extension, priority)
values
    ('requirements', 'txt', 'omni_python.install_requirements', 'omni_python', 1);

select pg_catalog.pg_extension_config_dump('auxiliary_tools', '');

create table procs as
    select *
    from
        pg_proc
    limit 0;

create table policies as
    select *
    from
        pg_policy
    limit 0;

create table class as
    select *
    from
        pg_class
    limit 0;
create function load_from_fs(fs anyelement, path text default '') returns setof text
    language plpgsql
as
$$
declare
    rec record;
begin
    -- Procs
    if not exists(select
                  from
                      pg_class
                  where
                      relname = '_omni_schema_pg_proc' and
                      relkind = 'r' and
                      relpersistence = 't') then
        create temporary table if not exists _omni_schema_pg_proc on commit drop as
            select * from pg_proc;
    end if;
    for rec in select * from omni_schema.procs
        loop
            execute format('drop function if exists %s', rec.oid::regprocedure);
        end loop;
    delete from omni_schema.procs;
    -- Policies
    if not exists(select
                  from
                      pg_class
                  where
                      relname = '_omni_schema_pg_policy' and
                      relkind = 'r' and
                      relpersistence = 't') then
        create temporary table if not exists _omni_schema_pg_policy on commit drop as
            select * from pg_policy;
    end if;
    for rec in select
                   policies.*,
                   pg_class.relname
               from
                   omni_schema.policies
                   inner join pg_class on pg_class.oid = policies.polrelid
        loop
            execute format('drop policy if exists %I on %I', rec.polname, rec.relname);
        end loop;
    delete from omni_schema.policies;
    -- Supported relations
    if not exists(select
                  from
                      pg_class
                  where
                      relname = '_omni_schema_pg_class' and
                      relkind = 'r' and
                      relpersistence = 't') then
        create temporary table if not exists _omni_schema_pg_class on commit drop as
            select * from pg_class;
    end if;
    for rec in select
                   class.*,
                   pg_namespace.nspname
               from
                   omni_schema.class
                   inner join pg_namespace on pg_namespace.oid = class.relnamespace
        loop
            if rec.relkind = 'v' then
                execute format('drop view if exists %I.%I cascade', rec.nspname, rec.relname);
            end if;
        end loop;
    delete from omni_schema.class;
    -- Execute
    for rec in select
                   case when path = '' then '' else path || '/' end || name as name,
                   omni_vfs.read(fs, path || '/' || name)                   as code,
                   language,
                   extension,
                   file_processor,
                   file_processor_extension,
                   processor,
                   processor_extension
               from
                   omni_vfs.list_recursively(fs, path, max => 10000) files
                   left join omni_schema.languages on files.name like concat('%', languages.file_extension)
                   left join omni_schema.auxiliary_tools
                             on files.name like concat(coalesce(auxiliary_tools.filename_stem, '%'), '.',
                                                       auxiliary_tools.filename_extension)
               where
                   auxiliary_tools.id is not null or
                   languages.id is not null
               order by coalesce(auxiliary_tools.priority, languages.priority) desc, name
        loop
            declare
                rec_code text;
            begin
                rec_code := convert_from(rec.code, 'utf-8');
                if rec.language is null and rec.processor is not null then
                    if not exists(select from pg_extension where extname = rec.processor_extension) then
                        raise notice 'Extension % required for auxiliary tool (required for %) is not installed', rec.processor_extension, rec.name;
                    else
                        execute format('select %s(%L::text)', rec.processor, rec_code);
                    end if;
                    continue;
                end if;
                if rec_code ~ 'omni_schema\[\[ignore\]\]' then
                    -- Ignore the file
                    continue;
                end if;
                if rec.language = 'sql' then
                    execute rec_code;
                else
                    if rec.language is null then
                        -- Ignore file
                        continue;
                    end if;
                    -- Check if the language is available
                    if not exists(select from pg_extension where extname = rec.extension) then
                        raise notice 'Extension % required for language % (required for %) is not installed', rec.extension, rec.language, rec.name;
                        -- Don't include it in the list of loaded files
                        continue;
                    else
                        -- Prepare and execute the SQL create function construct
                        declare
                            sql_snippet text;
                            regprocs    regprocedure[];
                        begin
                            if rec.file_processor is not null then
                                if (rec.file_processor_extension is not null and
                                    exists(select from pg_extension where extname = rec.file_processor_extension)) or
                                   rec.file_processor_extension is null then
                                    -- Can use the file processor
                                    execute format(
                                            'select array_agg(processor) from %s(%L::text, filename => %L::text, replace => true, fs => %L::%s) processor',
                                            rec.file_processor,
                                            rec_code, rec.name, fs::text, pg_typeof(fs)) into regprocs;
                                    if array_length(regprocs, 1) > 0 then
                                        return next rec.name;
                                        continue;
                                    end if;
                                end if;
                            end if;
                            if rec_code ~ 'SQL\[\[.*\]\]' then
                                sql_snippet := format('%s language %I as %L',
                                                      substring(rec_code from 'SQL\[\[(.*?)\]\]'), rec.language,
                                                      rec_code);
                                execute sql_snippet;
                            end if;
                        end;
                    end if;
                end if;
                return next rec.name;
            end;
        end loop;
    -- New procs
    insert
    into
        omni_schema.procs (select * from pg_proc except select * from _omni_schema_pg_proc);
    -- New policies
    insert
    into
        omni_schema.policies (select * from pg_policy except select * from _omni_schema_pg_policy);
    -- New supported relations
    insert
    into
        omni_schema.class (select * from pg_class except select * from _omni_schema_pg_class);
    return;
end ;
$$;
create table migrations
(
    id         integer primary key generated always as identity,
    name       text      not null,
    migration  text      not null,
    applied_at timestamp not null default now()
);

select pg_catalog.pg_extension_config_dump('migrations', '');
select pg_catalog.pg_extension_config_dump('migrations_id_seq', '');
create function migrate_from_fs(fs anyelement, path text default '') returns setof text
    language plpgsql
as
$$
declare
    rec record;
begin
    for rec in select
                   case when path = '' then '' else path || '/' end || files.name      as name,
                   convert_from(omni_vfs.read(fs, path || '/' || files.name), 'utf-8') as code
               from
                   omni_vfs.list_recursively(fs, path, max => 10000) as files
                   left join omni_schema.migrations
                             on migrations.name = (case when path = '' then '' else path || '/' end || files.name)
               where
                   files.name like '%.sql' and
                   migrations.name is null
               order by substring(files.name from '[0-9]+')::int asc
        loop
            execute rec.code;
            return next rec.name;
            insert
            into
                omni_schema.migrations (name, migration)
            values (rec.name, rec.code);
        end loop;
    return;
end;
$$;
create function assemble_schema(conn_info text, fs anyelement, path text default '')
    returns table
            (
                execution_position  int,
                migration_filename  text,
                migration_statement text,
                execution_error     text
            )
    language plpgsql
as
$$
declare
    rec                             record;
    remote_table_fs                 omni_vfs.table_fs;
    remote_table_fs_name            text    = 'omni_schema_migration_fs';
    current_filename                text;
    successful_statement_executions int     = 0;
    error_message                   text;
    error_detail text;
    try_again                       boolean = false;
begin
    -- statement execution status
    create temp table omni_schema_execution_status
    (
        id                       bigserial primary key,
        filepath                 text    not null,
        language                 text,
        extension                text,
        file_processor           text,
        file_processor_extension text,
        processor                text,
        processor_extension      text,
        code                     text    not null,
        -- order of successful execution
        execution_number         int,
        -- whether statement is executed successfully
        execution_successful     boolean not null default false,
        -- error for last execution of the statement
        last_execution_error     text
    ) on commit drop;

    <<top>>
    begin
        -- conn for setting up table_fs in remote database
        perform dblink_connect(remote_table_fs_name, conn_info);

        -- prepare necessary extensions in the remote database
        for rec in select distinct extension
                   from omni_schema.languages
                   union
                   select distinct file_processor_extension as extension
                   from omni_schema.languages
                   union
                   select distinct processor_extension as extension
                   from omni_schema.auxiliary_tools
            loop
                if exists (select from pg_extension where extname = rec.extension) then
                    perform dblink(remote_table_fs_name,
                                   format('create extension if not exists %s cascade', rec.extension));
                end if;
            end loop;

        perform dblink(remote_table_fs_name, 'create extension if not exists omni_vfs cascade');
        -- create table_fs in remote database
        select (f).id
        from dblink(remote_table_fs_name,
                    format('select omni_vfs.table_fs(%L)', remote_table_fs_name)) as t(f omni_vfs.table_fs)
        into remote_table_fs;

        -- read all files recognised by omni_schema.languages and omni_schema.auxiliary_tools
        for rec in
            select path || '/' || name                                           as name,
                   convert_from(omni_vfs.read(fs, path || '/' || name), 'UTF-8') as code,
                   language,
                   extension,
                   file_processor,
                   file_processor_extension,
                   processor,
                   processor_extension
            from omni_vfs.list_recursively(fs, path) files
                     left join omni_schema.languages on files.name like concat('%', languages.file_extension)
                     left join omni_schema.auxiliary_tools
                               on files.name like concat(coalesce(auxiliary_tools.filename_stem, '%'), '.',
                                                         auxiliary_tools.filename_extension)
            where (auxiliary_tools.id is not null or
                   languages.id is not null)
              and kind = 'file'
            order by coalesce(auxiliary_tools.priority, languages.priority) desc, name
            loop
                -- create table_fs entry for the file along with it's data
                perform dblink(remote_table_fs_name,
                               format(
                                       'insert into omni_vfs.table_fs_files(filesystem_id, filename, kind) values (%s, %L, ''file'')',
                                       remote_table_fs.id, rec.name));
                perform dblink(remote_table_fs_name,
                               format(
                                       'insert into omni_vfs.table_fs_file_data (file_id, data) values
                                          (
                                            omni_vfs.table_fs_file_id(omni_vfs.table_fs(%L), %L),
                                            %L
                                          )',
                                       remote_table_fs_name, rec.name, convert_to(rec.code, 'utf-8')));

                if rec.code ~ 'omni_schema\[\[ignore\]\]' then
                    -- Ignore the file
                    continue;
                end if;
                if rec.language = 'sql' then
                    -- process sql language files statement by statement
                    declare
                        code text := convert_from(omni_vfs.read(fs, rec.name), 'UTF-8');
                        code_rec record;
                    begin
                        for code_rec in select source as current_statement, line, col
                                                 from
                                                     omni_sql.raw_statements(code::cstring)
                            loop
                                raise notice '%', json_build_object('type', 'info', 'message', 'Executing', 'code',
                                                                    code_rec.current_statement,
                                                                    'file', rec.name, 'line', code_rec.line, 'col',
                                                                    code_rec.col);
                                insert
                                into omni_schema_execution_status(filepath, language, code)
                                values (rec.name, rec.language, code_rec.current_statement);
                            end loop;
                    exception
                        when syntax_error then
                            raise notice '%', json_build_object('type', 'error', 'message', sqlerrm,
                                                                'file', rec.name);
                            insert
                            into omni_schema_execution_status(filepath, language, code, last_execution_error)
                            values (rec.name, rec.language, code, sqlerrm);

                            exit top;
                    end;
                else
                    -- process the whole file at once for other languages
                    insert
                    into omni_schema_execution_status(filepath, language, extension, file_processor,
                                                      file_processor_extension, processor, processor_extension, code)
                    values (rec.name, rec.language, rec.extension, rec.file_processor, rec.file_processor_extension,
                            rec.processor, rec.processor_extension, rec.code);
                end if;
            end loop;
        for rec in
            select distinct filepath from omni_schema_execution_status
            loop
                /*
                    create separate connections for each file to restrict any 'SET' statement
                    effects to the file which contains it
                */
                perform dblink_connect(rec.filepath, conn_info);
            end loop;
        -- execute until all statements are successfully executed or no more executions are successful
        while true
            loop
                <<file>>
                for current_filename in select distinct filepath from omni_schema_execution_status order by filepath
                    loop
                        <<statement>>
                        for rec in select *
                                   from omni_schema_execution_status
                                   where execution_successful = false
                                     and filepath = current_filename
                                   order by id
                            loop
                                declare
                                    _filepath text;
                                    _code     text;
                                begin
                                    if rec.language is null and rec.processor is not null then
                                        if not exists(select *
                                                      from dblink(rec.filepath,
                                                                  format(
                                                                          'select true from pg_extension where extname = %L',
                                                                          rec.processor_extension)) as t(b boolean)) then
                                            raise exception 'Extension % required for auxiliary tool (required for %) is not installed', rec.processor_extension, rec.filepath;
                                        else
                                            perform *
                                            from dblink(rec.filepath,
                                                        format('with cte as (%s) select true from cte',
                                                               format('select %s(%L::text)', rec.processor, rec.code))) as t(b boolean);
                                        end if;
                                    elsif rec.language = 'sql' then
                                        -- normalize sql statement like removing whitespace
                                        rec.code = omni_sql.statement_in(rec.code::cstring);
                                        if omni_sql.statement_type(rec.code::omni_sql.statement) = 'SelectStmt' then
                                            /*
                                                SELECT statement result set can be of any shape(number of columns, their types)
                                                so wrap select statements in cte to give it consistent shape and avoid the
                                                ERROR:  function returning record called in context that cannot accept type record
                                            */
                                            perform *
                                            from dblink(rec.filepath,
                                                        format('with cte as (%s) select true from cte', rec.code)) as t(b boolean);
                                        else
                                            perform dblink(rec.filepath, rec.code);
                                        end if;
                                    else
                                        if not exists(select *
                                                      from dblink(rec.filepath,
                                                                  format(
                                                                          'select true from pg_extension where extname = %L',
                                                                          rec.extension)) as t(b boolean)) then
                                            raise exception 'Extension % required for language % (required for %) is not installed', rec.extension, rec.language, rec.filepath;
                                        else
                                            -- Prepare and execute the SQL create function construct
                                            if rec.file_processor is not null then
                                                if (rec.file_processor_extension is not null and
                                                    exists(select *
                                                           from dblink(rec.filepath,
                                                                       format(
                                                                               'select true from pg_extension where extname = %L',
                                                                               rec.file_processor_extension)) as t(b boolean))) or
                                                   rec.file_processor_extension is null then
                                                    -- Can use the file processor
                                                    perform *
                                                    from dblink(rec.filepath, format(
                                                            'with cte as (select %s(%L::text, filename => %L::text, replace => true, fs => %L::%s)) select true from cte',
                                                            rec.file_processor,
                                                            rec.code, rec.filepath, remote_table_fs::text,
                                                            pg_typeof(remote_table_fs))) as t(b boolean);
                                                end if;
                                            end if;
                                            if rec.code ~ 'SQL\[\[.*\]\]' then
                                                perform dblink(rec.filepath, format('%s language %I as %L',
                                                                                    substring(rec.code from 'SQL\[\[(.*?)\]\]'),
                                                                                    rec.language,
                                                                                    rec.code));
                                            end if;
                                        end if;
                                    end if;
                                    -- update statement execution status
                                    successful_statement_executions = successful_statement_executions + 1;
                                    update omni_schema_execution_status
                                    set execution_number     = successful_statement_executions,
                                        execution_successful = true,
                                        last_execution_error = null
                                    where id = rec.id
                                    returning filepath, code into _filepath, _code;
                                    raise notice '%', json_build_object('type', 'info', 'message',
                                                                        'Completed', 'code', _code,
                                                                        'file', _filepath);
                                    -- at least one succeeded, worth trying again
                                    try_again = true;
                                exception
                                    when others then
                                        get stacked diagnostics error_message = message_text, error_detail = pg_exception_detail;
                                        update omni_schema_execution_status
                                        set last_execution_error = error_message
                                        where id = rec.id
                                        returning filepath, code into _filepath, _code;
                                        raise notice '%', json_build_object('type', 'error', 'message',
                                                                            sqlerrm, 'detail', error_detail, 'code',
                                                                            _code,
                                                                            'file', _filepath);
                                        -- go to next file if statement execution fails to preserve serial execution of statements in a file
                                        continue file;
                                end;
                            end loop statement;
                    end loop file;
                if try_again = true then
                    try_again = false;
                else
                    exit;
                end if;
            end loop;

        for rec in
            select distinct filepath from omni_schema_execution_status
            loop
                perform dblink_disconnect(rec.filepath);
            end loop;
    end;
    -- close the connections set up at the beginning
    perform dblink_disconnect(remote_table_fs_name);

    return query select execution_number,
                        filepath,
                        code,
                        last_execution_error
                 from omni_schema_execution_status
                 order by execution_number;
end;
$$;
