create type unit;

create function unit_in(cstring) returns unit
as
'MODULE_PATHNAME' language c immutable;

create function unit_out(unit) returns cstring
as
'MODULE_PATHNAME' language c immutable;

create function unit_recv(internal) returns unit
as
'MODULE_PATHNAME' language c immutable;

create function unit_send(unit) returns bytea
as
'MODULE_PATHNAME' language c immutable;


create type unit
(
    input = unit_in,
    output = unit_out,
    send = unit_send,
    receive = unit_recv,
    internallength = 1,
    alignment = char,
    storage = plain,
    passedbyvalue
);

create function unit() returns unit
as
'MODULE_PATHNAME' language c immutable;
create table sum_types
(
    typ      regtype primary key unique,
    variants regtype[] not null
);

create or replace function sum_type_unique_variants_trigger_func()
    returns trigger as
$$
declare
    duplicate_count integer;
begin
    select
        count(*) - count(distinct element)
    into duplicate_count
    from
        unnest(new.variants) as element;

    if duplicate_count > 0 then
        raise exception 'Sum types can not contain duplicate variants';
    end if;
    return new;
end;
$$ language plpgsql;

create trigger sum_type_unique_variants_trigger
    before insert or update
    on sum_types
    for each row
execute function sum_type_unique_variants_trigger_func();

create function sum_type(name name,
                         variadic variants regtype[]
) returns regtype
as
'MODULE_PATHNAME',
'sum_type'
    language c;

create function add_variant(sum_type regtype, variant regtype) returns void
as
'MODULE_PATHNAME',
'add_variant'
    language c;

create function variant(v anycompatible) returns regtype as
'MODULE_PATHNAME',
'sum_variant' language c;
create table function_signature_types
(
    typ       regtype primary key unique,
    arguments regtype[] not null,
    return    regtype   not null
);

select pg_catalog.pg_extension_config_dump('function_signature_types', '');

create function function_signature_type_of(name name, func text)
    returns regtype
    language plpgsql
as
$$
declare
    result regtype;
    proc   regprocedure;
begin
    set local search_path to omni_polyfill, '$user', public, pg_catalog;
    begin
        proc = func::regprocedure;
    exception
        when others then
            proc = func::regproc;
    end;
    select omni_types.function_signature_type(name, variadic
                                              trim_array(
                                                      array_append(pg_proc.proargtypes::oid[], pg_proc.prorettype)::regtype[],
                                                      0))
    from pg_proc
    where oid = proc
    into result;
    return result;
end;
$$;

create function function_signature_type(name name,
                                        variadic signature regtype[]
) returns regtype
    language plpgsql
as
$$
declare
    ret_type           regtype;
    omni_types_library text;
    rec record;
begin
    set local search_path to omni_polyfill, '$user', public, pg_catalog;

    for rec in select typ, arguments, return from omni_types.function_signature_types where typ::name = name
        loop
            declare
                existing_sig regtype[];
            begin
                existing_sig := trim_array(array_append(rec.arguments, rec.return), 0);
                if not array_eq(existing_sig, signature) then
                    raise exception 'cannot redefine % as its signature % does not match %', name, existing_sig, signature;
                else
                    return rec.typ;
                end if;
            end;
        end loop;

    -- Resolve omni_types module
    select probin
    into omni_types_library
    from pg_proc
             inner join pg_namespace on pg_proc.pronamespace = pg_namespace.oid and pg_namespace.nspname = 'omni_types'
    where proname = 'unit_in';

    -- Define the shell of the type
    execute format('create type %I;', name);

    -- Make sure the commands below don't produce these messages:
    -- NOTICE:  return type a is only a shell
    -- NOTICE:  argument type a is only a shell
    -- NOTICE:  argument type a is only a shell
    -- NOTICE:  return type a is only a shell
    --
    -- (they are annoying and unnecessary)
    set local client_min_messages = warning;

    -- Define in/out send/recv before we actually define the type

    execute format('create function %I(cstring) returns %I language c as %L, %L immutable strict', name || '_in',
                   name,
                   omni_types_library, 'function_signature_in');

    execute format('create function %I(%I) returns cstring language internal as %L immutable strict', name || '_out',
                   name,
                   'regprocedureout');

    execute format('create function %I(%I) returns bytea language internal as %L immutable',
                   name || '_send', name,
                   'regproceduresend');
    execute format('create function %I(internal) returns %I language internal as %L immutable',
                   name || '_recv', name,
                   'regprocedurerecv');

    -- Resume notices disabled above
    reset client_min_messages;

    -- Finally, define the type
    execute format(
            'create type %I ( internallength = %s, input = %I, output = %I, send = %I, receive = %I, like = oid, storage = plain)',
            name, (select typlen from pg_type where typname = 'oid'), name || '_in', name || '_out', name || '_send',
            name || '_recv'
            );

    -- Register the type
    ret_type := signature[cardinality(signature)];
    insert into omni_types.function_signature_types (typ, arguments, return)
    values (name::regtype, trim_array(signature, 1), ret_type);

    -- Define casts out of the type
    execute format('create cast (%I as regproc) without function', name);
    execute format('create cast (%I as regprocedure) without function', name);
    execute format('create cast (%I as oid) without function', name);

    -- Define conformance test function
    execute format('create function %3$I(text) returns %1$I language c as %2$L, ''conforming_function''',
                   name,
                   omni_types_library, name || '_conforming_function');

    -- Define caller function
    execute format('create function %I(%I %s) returns %s language c as %L, %L', 'call_' || name, name,
                   case
                       when cardinality(signature) = 1 then ''
                       else ', ' || concat_ws(', ', variadic trim_array(signature, 1)) end,
                   ret_type, omni_types_library, 'invoke');

    return name::regtype;
end;
$$;
select pg_catalog.pg_extension_config_dump('sum_types', '');
